import sys
from configparser import ConfigParser
from os import mkdir
from pathlib import Path

import appdirs

config_file = (
    Path(appdirs.user_config_dir("ttun-client", "ttun", roaming=True)) / "config.ini"
)

config = ConfigParser()
config["server"] = {
    "hostname": "",
    "using_ssl": "yes",
}

if not config_file.parent.exists():
    mkdir(config_file.parent)
else:
    config.read(filenames=config_file)

with open(config_file, "w") as f:
    config.write(f)

SERVER_HOSTNAME = config["server"].get("hostname")
SERVER_USING_SSL = config["server"].get("using_ssl").lower() in ["yes", "true"]

try:
    assert SERVER_HOSTNAME != ""
except AssertionError:
    print(f"Please add a valid server hostname in {config_file}")
    sys.exit(1)



from typing import Optional
from typing import TypedDict


class Message(TypedDict):
    type: str


class RequestData(TypedDict):
    method: str
    path: str
    headers: list[tuple[str, str]]
    body: Optional[str]


class ResponseData(TypedDict):
    status: int
    headers: list[tuple[str, str]]
    body: Optional[str]


class Config(TypedDict):
    url: str



import asyncio
import json
from base64 import b64decode
from base64 import b64encode
from datetime import datetime
from time import perf_counter
from typing import Awaitable
from typing import Callable
from typing import Coroutine
from typing import Optional
from uuid import uuid4

import websockets
from aiohttp import ClientSession
from aiohttp import DummyCookieJar
from websockets import WebSocketClientProtocol
from websockets.exceptions import ConnectionClosed

from ttun.pubsub import PubSub
from ttun.types import Config
from ttun.types import RequestData
from ttun.types import ResponseData


class Client:
    def __init__(self, port: int, server: str, subdomain: str = None):
        self.port = port
        self.server = server
        self.subdomain = subdomain

        self.config: Optional[Config] = None
        self.connection: WebSocketClientProtocol = None

    async def send(self, data: dict):
        await self.connection.send(json.dumps(data))

    async def receive(self) -> dict:
        return json.loads(await self.connection.recv())

    @staticmethod
    def loop(sleep: int = None):
        async def wrapper(callback: Callable[[], Coroutine]):
            while True:
                try:
                    await callback()

                    if sleep is not None:
                        await asyncio.sleep(sleep)

                except ConnectionClosed:
                    break

        return wrapper

    async def connect(self) -> WebSocketClientProtocol:
        self.connection = await websockets.connect(f"{self.server}/tunnel/")

        await self.send({"subdomain": self.subdomain})

        self.config = await self.receive()

        if self.connection.open:
            return self.connection

    async def handle_messages(self):
        while True:
            try:
                request: RequestData = await self.receive()
                await self.proxyRequest(
                    request=request, on_response=lambda response: self.send(response)
                )

            except ConnectionClosed:
                break

    async def proxyRequest(
        self,
        request: RequestData,
        on_response: Callable[[ResponseData], Awaitable] = None,
    ):
        async with ClientSession(cookie_jar=DummyCookieJar()) as session:
            request_id = uuid4()
            await PubSub.publish(
                {
                    "type": "request",
                    "payload": {
                        "id": request_id.hex,
                        "timestamp": datetime.now().isoformat(),
                        **request,
                    },
                }
            )

            start = perf_counter()
            response = await session.request(
                method=request["method"],
                url=f'http://localhost:{self.port}{request["path"]}',
                headers=request["headers"],
                data=b64decode(request["body"].encode()),
                allow_redirects=False,
            )
            end = perf_counter()

            response_data = ResponseData(
                status=response.status,
                headers=[
                    (key, value)
                    for key, value in response.headers.items()
                    if key.lower() not in ["transfer-encoding", "content-encoding"]
                ],
                body=b64encode(await response.read()).decode(),
            )

            if on_response is not None:
                await on_response(response_data)

            await PubSub.publish(
                {
                    "type": "response",
                    "payload": {
                        "id": request_id.hex,
                        "timing": end - start,
                        **response_data,
                    },
                }
            )



import asyncio
from contextlib import contextmanager
from typing import Any
from typing import Generator
from typing import Iterator


class PubSub:
    _instance = None

    def __init__(self):
        self.queues: list[asyncio.Queue] = []
        self._history = []

    @classmethod
    def instance(cls) -> "PubSub":
        if cls._instance == None:
            cls._instance = cls()
        return cls._instance

    @classmethod
    @contextmanager
    def subscribe(cls) -> Iterator[asyncio.Queue]:
        queue = asyncio.Queue()

        instance = cls.instance()
        instance.queues.append(queue)

        try:
            yield queue
        finally:
            cls.unsubscribe(queue)

    @classmethod
    def unsubscribe(cls, queue: asyncio.Queue):
        instance = cls.instance()
        instance.queues.remove(queue)
        del queue

    @classmethod
    @property
    def history(cls):
        instance = cls.instance()
        return instance._history

    @classmethod
    async def publish(cls, msg: Any):
        instance = cls.instance()

        instance._history.append(msg)
        for queue in instance.queues:
            await queue.put(msg)



from importlib import resources
from pathlib import Path
from typing import Awaitable
from typing import Callable
from typing import Optional

from aiohttp import web

from ttun.pubsub import PubSub
from ttun.types import Config
from ttun.types import RequestData

BASE_DIR = Path(__file__).resolve().parent


class Server:
    def __init__(
        self,
        config: Config,
        on_resend: Callable[[RequestData], Awaitable],
        on_started: Callable[["Server"], None],
    ):
        self.port = 4040

        self.config = {**config, "assets": "/assets/"}

        self.on_resend = on_resend
        self.on_started = lambda *args, **kwargs: on_started(self)

        self.app = web.Application()
        with resources.path(__package__, "staticfiles") as staticfiles:
            self.app.add_routes(
                [
                    web.get("/inspect/", self.inspect_socket),
                    web.get("/config/", self.get_config),
                    web.post("/resend/", self.resend),
                    web.get("/", self.root),
                    web.static("/", staticfiles),
                ]
            )

    async def run(self):
        while True:
            try:
                await web._run_app(self.app, port=self.port, print=self.on_started)
            except OSError:
                self.port += 1

    async def root(self, request: web.Request):
        with resources.path(__package__, "staticfiles") as staticfiles:
            return web.Response(
                body=resources.read_text(f"{__package__}.staticfiles", f"index.html"),
                content_type="text/html",
            )

    async def get_config(self, request: web.Request):
        return web.json_response(
            self.config, headers={"Access-Control-Allow-Origin": "*"}
        )

    async def resend(self, request: web.Request):
        await self.on_resend(await request.json())
        return web.json_response()

    async def inspect_socket(self, request: web.Request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        await ws.send_json({"type": "historic", "payload": PubSub.history})

        with PubSub.subscribe() as subscription:
            while message := await subscription.get():
                await ws.send_json(message, False)

        return ws



import asyncio
from argparse import ArgumentDefaultsHelpFormatter
from argparse import ArgumentParser
from asyncio import FIRST_EXCEPTION
from asyncio.exceptions import CancelledError
from asyncio.exceptions import TimeoutError

from ttun.client import Client
from ttun.inspect_server import Server
from ttun.settings import SERVER_HOSTNAME
from ttun.settings import SERVER_USING_SSL

inspect_queue = asyncio.Queue()


def main():
    parser = ArgumentParser(prog="ttun", formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("port", help="The local port to expose")
    parser.add_argument(
        "--server",
        default=f'{"wss" if SERVER_USING_SSL else "ws"}://{SERVER_HOSTNAME}',
        help="The hostname of the ttun server",
    )
    parser.add_argument(
        "-s",
        "--subdomain",
        default=None,
        help="The subdomain of the ttun tunnel",
    )
    args = parser.parse_args()

    client = Client(
        port=args.port,
        subdomain=args.subdomain,
        server=args.server,
    )

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    loop.run_until_complete(client.connect())

    def print_info(server: Server):
        print("Tunnel created:")
        print(f'{client.config["url"]} -> http://localhost:{args.port}')
        print("")
        print(f"Inspect requests:")
        print(f"http://localhost:{server.port}")

    server = Server(
        config=client.config,
        on_resend=client.proxyRequest,
        on_started=print_info,
    )

    tasks = {loop.create_task(client.handle_messages()), loop.create_task(server.run())}

    try:
        loop.run_until_complete(asyncio.wait(tasks, return_when=FIRST_EXCEPTION))
    except (CancelledError, TimeoutError):
        for task in tasks:
            task.cancel()
        loop.close()
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()
